#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lidl.h"
#include "datastructs.h"

void qTreeInsert(QTree *t, Object *o);
void subInsert(QTree *t, Object *o);
int qTreeOverflow(QTree *t);
void subdivideQTree(QTree *t);
void flattenQTree(QTree *t);
void expandToCover(QTree **treep, Pos pos);

QTree * addQTree(QTree *parent, Pos pos, unsigned int width)
{
    QTree *t;

    t = (QTree *)malloc(sizeof(QTree));
    memset(t, 0, sizeof(QTree));

    t->content = addList();
    t->pos = pos;
    t->width = width;
    t->parent = parent;

    return t;
}

void delQTree(QTree *t)
{
    if(t->content->first)
	die("Can't delete non-empty qtree");
    if(t->nw)
	delQTree(t->nw);
    if(t->ne)
	delQTree(t->ne);
    if(t->sw)
	delQTree(t->sw);
    if(t->se)
	delQTree(t->se);
    if(t->parent)
    {
       if(t->parent->nw == t)
	  t->parent->nw = NULL;
       else if(t->parent->ne == t)
	  t->parent->ne = NULL;
       else if(t->parent->sw == t)
	  t->parent->sw = NULL;
       else if(t->parent->se == t)
	  t->parent->se = NULL;
    }
    free(t);
}

QTree * addQRoot(void)
{
    return addQTree(NULL, mkPos(-1,-1), 2);
}

void addQTreeSize(QTree *t, int num)
{
    while(t)
    {
	t->size += num;
	t = t->parent;
    }
}

int outsideQTree(QTree *t, Pos pos)
{
    if(pos.x < t->pos.x || pos.y < t->pos.y)
	return -1;
    if(pos.x >= t->pos.x + (int)t->width || pos.y >= t->pos.y + (int)t->width)
	return 1;
    return 0;
}

int qTreeOverflow(QTree *t)
{
    if(t->size < QTREE_THRESHOLD || t->width % 2)
	return FALSE;
    return TRUE;
}

/* subdivides a node and distributes it's contents to the subnodes */
void subdivideQTree(QTree *t)
{
    Pos pos;
    int width = t->width / 2;
    unsigned int oldsize = t->content->size;

    if(!t->nw)
	t->nw = addQTree(t, t->pos, width);
    if(!t->ne)
    {
	pos.x = t->pos.x + width;
	pos.y = t->pos.y;
	t->ne = addQTree(t, pos, width);
    }
    if(!t->sw)
    {
	pos.x = t->pos.x;
	pos.y = t->pos.y + width;
	t->sw = addQTree(t, pos, width);
    }
    if(!t->se)
    {
	pos.x = t->pos.x + width;
	pos.y = t->pos.y + width;
	t->se = addQTree(t, pos, width);
    }

    while(t->content->first)
    {
	subInsert(t, t->content->first->p);
	delListNode(t->content->first);
    }

    addQTreeSize(t, -oldsize);
}

/* packs the contents of a node recursively into itself */
void flattenQTree(QTree *t)
{
    int i;
    ListNode *ln;
    QTree *st;

    for(i = 0; i < 4; i++)
    {
	switch(i)
	{
	case 0: st = t->nw; break;
	case 1: st = t->ne; break;
	case 2: st = t->sw; break;
	case 3: st = t->se; break;
	}
	if(st)
	{
	    flattenQTree(st);
	    while(st->content->first)
	    {
		ln = addListNode(t->content);
		ln->p = st->content->first->p;
		delListNode(st->content->first);
	    }
	    delQTree(st);
	}
    }
}

/*
  expandToCover() scales the rootnode with 2 and creates subnodes,
  reinserts the old roots subnodes into these and frees the old root
  until the position is located inside the rootnode.
 */
void expandToCover(QTree **treep, Pos pos)
{
    QTree *root = *treep;

    while(outsideQTree(root, pos))
    {
	int ow = root->width;
	Pos npos = posSub(root->pos, mkPos(ow/2,ow/2));

	if(root->content->first)
	{
	    root->pos = npos;
	    root->width = ow*2;
	}
	else
	{
	    QTree *t, *newroot = addQTree(NULL, npos, ow*2);

	    if(root->nw)
	    {
		t = addQTree(newroot, newroot->pos, ow);
		t->se = root->nw;
		t->se->parent = t;
		newroot->nw = t;
		t->size = root->nw->size;
		swapList(&root->nw->content, &t->content);
		if(!qTreeOverflow(t))
		    flattenQTree(t);
	    }
	    if(root->ne)
	    {
		t = addQTree(newroot, posAdd(newroot->pos, mkPos(ow,0)), ow);
		t->sw = root->ne;
		t->sw->parent = t;
		newroot->ne = t;
		t->size = root->ne->size;
		swapList(&root->ne->content, &t->content);
		if(!qTreeOverflow(t))
		    flattenQTree(t);
	    }
	    if(root->sw)
	    {
		t = addQTree(newroot, posAdd(newroot->pos, mkPos(0,ow)), ow);
		t->ne = root->sw;
		t->ne->parent = t;
		newroot->sw = t;
		t->size = root->sw->size;
		swapList(&root->sw->content, &t->content);
		if(!qTreeOverflow(t))
		    flattenQTree(t);
	    }
	    if(root->se)
	    {
		t = addQTree(newroot, posAdd(newroot->pos, mkPos(ow,ow)), ow);
		t->nw = root->se;
		t->nw->parent = t;
		newroot->se = t;
		t->size = root->se->size;
		swapList(&root->se->content, &t->content);
		if(!qTreeOverflow(t))
		    flattenQTree(t);
	    }
	    root->nw = root->ne = root->sw = root->se = NULL;
	    delQTree(root);
	    root = newroot;
	}
    }

    *treep = root;
}

/* internal insert */
void qTreeInsert(QTree *t, Object *o)
{
    if(!qTreeOverflow(t))
    {
	addListNode(t->content)->p = o;
	addQTreeSize(t, 1);
	return;
    }
    subdivideQTree(t);
    subInsert(t, o);
}

/* this function assumes that all subnodes is present */
void subInsert(QTree *t, Object *o)
{
    if(!outsideQTree(t->nw, o->pos))
	qTreeInsert(t->nw, o);
    else if(!outsideQTree(t->ne, o->pos))
	qTreeInsert(t->ne, o);
    else if(!outsideQTree(t->sw, o->pos))
	qTreeInsert(t->sw, o);
    else if(!outsideQTree(t->se, o->pos))
	qTreeInsert(t->se, o);
    else
	die("subInsert given object outside qtree");
}

/* the insertfunction to be used by external functions */
void insertObject(QTree **tp, Object *o)
{
    expandToCover(tp, o->pos);
    qTreeInsert(*tp, o);
}

List * getObjectsAt(QTree *t, Pos pos)
{
    List *ret;
    ListNode *ln;

    if(!t || !t->size || outsideQTree(t, pos))
	return NULL;

    if(t->content->first)
    {
	ret = addList();
	for(ln = t->content->first; ln; ln = ln->next)
	    if(!posCmp(pos, ((Object *)ln->p)->pos))
		addListNode(ret)->p = ln->p;
	return ret;
    }

    if(t->nw && !outsideQTree(t->nw, pos))
	return getObjectsAt(t->nw, pos);
    else if(t->ne && !outsideQTree(t->ne, pos))
	return getObjectsAt(t->ne, pos);
    else if(t->sw && !outsideQTree(t->sw, pos))
	return getObjectsAt(t->sw, pos);
    else if(t->se && !outsideQTree(t->se, pos))
	return getObjectsAt(t->se, pos);

    return NULL;
}

Object * getObjectAt(QTree *t, Pos pos)
{
    List *l = getObjectsAt(t, pos);
    if(!l || !l->first)
	return NULL;
    return l->first->p;
}
