/* UI handling */

#include <stdio.h>
#include <stdarg.h>
#include <curses.h>

#include "lidl.h"
#include "ui.h"
#include "world.h"

WINDOW *conwin;  /* used for input/output */
WINDOW *mapwin;  /* suprice used for map drawing */
WINDOW *statwin; /* HP/MP etc */

int uiColours;

void uiDrawMap( Pos p, char type, int colourPair, int bold );

void uiPrint( char *fmt, ... )
{
    /* prints the string as the ui wishes */

    va_list ap;
    va_start( ap, fmt );
    vwprintw( conwin, fmt, ap );
    va_end( ap );
    wprintw( conwin, "\n" );
    uiUpdate();
}

void uiInputString( char *buf, size_t bufsize, const char *string )
{
    /* string will be the promt.
       will return a string */

    wprintw( conwin, "%s: ", string );
    uiUpdate();
    wgetnstr( conwin, buf, bufsize );
    uiUpdate();
}

enum {SCREEN_WIDTH = 60, SCREEN_HEIGHT = 13};

void uiInit( void )
{
    /* Used to initialize the ui specific stuff */
    
    initscr();
    if( ! has_colors() )
    {
        uiColours = FALSE;
    }
    else
    {
        uiColours = TRUE;
        start_color();
        init_pair( 0, COLOR_WHITE, COLOR_BLACK );   /* Default-ish */
        init_pair( 1, COLOR_YELLOW, COLOR_YELLOW );
    }

    conwin = newwin( 10, 80, 14, 0 ); /* lines, cols, y, x */

    mapwin = newwin( 14, 60, 0, 0 );
    statwin = newwin( 14, 20, 0, 60 );
    keypad( conwin, TRUE ); /* make stdscr the input */
    nonl();
    cbreak();
    echo();                 /* oooh yes local echo */
    scrollok( conwin, TRUE );
    mvwhline( mapwin, 13, 0, ACS_HLINE, 60 ); /* draw a line */
    mvwvline( statwin, 0, 0, ACS_VLINE, 13 ); /* one short so we might add a nice "bend" */
    mvwaddch( statwin, 13, 0, ACS_LLCORNER );
    mvwhline( statwin, 13, 1, ACS_HLINE, 19 );
    
    /* add the captions to stat window */
    mvwprintw( statwin, 0, 2, "HP: " );
    mvwprintw( statwin, 1, 2, "MP: " );
    mvwprintw( statwin, 2, 2, "STR: " );
    mvwprintw( statwin, 3, 2, "CON: " );
    mvwprintw( statwin, 4, 2, "DEX: " );
    mvwprintw( statwin, 5, 2, "SIZ: " );
    mvwprintw( statwin, 7, 2, "DLVL: " );
    mvwprintw( statwin, 8, 2, "Moves: " );
    mvwprintw( statwin, 9, 2, "Score: " );
#ifdef DEBUG
    mvwprintw( statwin, 10, 2, "Pos: " );
#endif
}

void uiStop( void )
{
    /* Some kind of ui's will need this */
    endwin();
}

void uiUpdate( void )
{
    /* used by the ui internaly for updates
       NOTE: there should not be a need to hail this externaly */
  
    wrefresh( mapwin );
    uiUpdateStats();
    wrefresh( statwin );
    wrefresh( conwin ); /* must be last to regain focus! */
}

void uiUpdateMap()
{
    /* draws the map */
  
    Pos p, pp = playerGetPosition();
    
    for( p.x = 0; p.x < SCREEN_WIDTH; p.x++ )
        for( p.y = 0; p.y < SCREEN_HEIGHT; p.y++ )
        {
            debugPrint( "UpdateMap %d, %d", p.x, p.y );
            switch( mapTypeAt( posAdd( p, mkPos( pp.x - SCREEN_WIDTH/2, pp.y - SCREEN_HEIGHT/2) ) ) )
            {
	    case worldERROR:
                uiDrawMap( p, 'E', 0, FALSE );
                break;
	    case worldEMPTY:
                uiDrawMap( p, '.', 0, FALSE );
                break;
            case worldWALL:
                uiDrawMap( p, '#', 1, FALSE );
                break;
	    case worldHOLE:
                uiDrawMap( p, '^', 0, FALSE ); /* find something better */
                break;
	    case mapUNSEEN:
                uiDrawMap( p, ' ', 0, FALSE );
                break;
	    case mapPLAYER:
                uiDrawMap( p, '@', 0, TRUE ); /* Ala nethack */
                break;
            }
	}

    uiUpdate(); /* update the ui */
}

void uiDrawMap( Pos p, char type, int colourPair, int bold )
{
    /* Used internaly by uiUpdateMap
       Note: will skip colours and boldness if we got no colour support */
    if( uiColours )
    {
        if( bold )
            wattron( mapwin, A_BOLD );
        wattron( mapwin, COLOR_PAIR( colourPair ));
        mvwaddch( mapwin, p.y, p.x, type );
        wattroff( mapwin, COLOR_PAIR( colourPair ));
        if( bold )
            wattroff( mapwin, A_BOLD );
    }
    else
        mvwaddch( mapwin, p.y, p.x, type );
}

int uiInputChar( void )
{
    /* A char will be read
       Note: This function is silent now
       Note: When we use the numpad we break some support for digit input :/ */

    int buf;

    noecho();
    buf = wgetch( conwin );
    echo();

    /* code to convert ui specific to game codes
       FIXME: don't use the numpad in "number mode" */
    switch( buf )
    {
    case '8':
        buf = uiMOVEN;
        break;
    case '2':
        buf = uiMOVES;
        break;
    case '4':
        buf = uiMOVEW;
        break;
    case '6':
        buf = uiMOVEE;
        break;
    case '7':
        buf = uiMOVENW;
        break;
    case '9':
        buf = uiMOVENE;
        break;
    case '1':
        buf = uiMOVESW;
        break;
    case '3':
        buf = uiMOVESE;
        break;
    }

    debugPrint( "buf = %d", buf );

    return buf;
}

void uiInfoBox( char *fmt, ... )
{
    /* creates and shows a info box
       In this implemention it might be dismissed with 'q'
       And it offers 20x40 chars
	 
       TODO: this code would not get sad if it got somewhat butified
	 
       TODO(2): This code should support several pages, and the window
       should adapt it self to the screen */

    WINDOW *infowin;
    WINDOW *infoborderwin;
    va_list ap;

    infowin = newwin( 20, 40, 2, 18 );
    infoborderwin = newwin( 22, 42, 1, 17 );

    va_start( ap, fmt );
    vwprintw( infowin, fmt, ap );
    va_end( ap );

    box( infoborderwin, ACS_VLINE, ACS_HLINE );
    uiUpdate();
    noecho();
    wrefresh( infoborderwin );
    wrefresh( infowin ); /* we are not a standard window */
  
    while( wgetch( infowin ) != 'q' ); /* wait until the signal! */
    delwin( infowin ); /* when done remove the window */
    delwin( infoborderwin );
    /* restore the screen */
    echo();
    redrawwin( conwin );
    redrawwin( statwin );
    redrawwin( mapwin );
    uiUpdate();
}

void uiUpdateStats( void )
{
    /* NOTE: This is a ui specific function and should never be called
       on by alien code.
       Its here just to clean up the code */

#ifdef DEBUG
    Pos p = playerGetPosition();
#endif

    mvwprintw( statwin, 0, 7, "%d/%d", playerGetHp(), playerGetHpMax() );
    mvwprintw( statwin, 8, 9, "%d", playerGetMoves() );
#ifdef DEBUG
    mvwprintw( statwin, 10, 9, "%d,%d", p.x, p.y );
#endif
}
